import numpy as np
import pandas as pd
from ..TimeSeriesAnalysis import CovMatrix


class InverseVariance():

    """
    Inverse Variance Portfolio Optimization

    Parameters
    ----------
    AssetPrices : pd.DataFrame

    RiskMetric : :py:class:`str`, optional
        Available options are: "Variance", "Semivariance". The default is "Variance".

    Period : :py:class:`str`, optional
        The time period used to calculate the weights. The default is "days".


    Functions
    ---------
    Optimize() : Returns the optimal weights using inverse variance algorithm.
    """

    def __init__(self, AssetPrices: pd.DataFrame, RiskMetrics: str = "Variance", Period: str = "days"):

        # Check if the asset prices are valid
        if AssetPrices is not None:
            if not isinstance(AssetPrices, pd.DataFrame):
                raise ValueError(
                    "Asset Prices must be a Pandas DataFrame!")
            if not isinstance(AssetPrices.index, pd.DatetimeIndex):
                raise ValueError(
                    "Asset Prices must be a Pandas DataFrame index by datetime!")

        self.AssetNames = AssetPrices.columns

        if RiskMetrics.lower() in ["semi", "semivar", "semivariance"]:
            self.risk_matrix = CovMatrix.SampleSemi(AssetPrices)

        elif RiskMetrics.lower() in ["cov", "covariance", "variance", "var"]:
            self.risk_matrix = CovMatrix.Sample(AssetPrices)

        elif RiskMetrics.lower() in ["lw", "letoid", "letoidwolf"]:
            self.risk_matrix = CovMatrix.LedoitWolf(AssetPrices)

        else:
            raise ValueError(
                "RiskMetrics must be one 'semivariance' or 'variance'")

    def Optimize(self, Decimals: int = 4):
        """
        Calculate asset weights using inverse variance algorithm.

        Parameters
        ----------

        Decimals : :py:class:`int`, optional
            The number of decimals to round the weights. The default is 4.

        Returns
        -------
        Portfolio : :py:class:`pandas.DataFrame`
        """

        Portifolio = 1. / np.diag(self.risk_matrix)
        Portifolio = Portifolio.round(Decimals)
        Portifolio /= Portifolio.sum()

        Portifolio[Portifolio == 0] = 'nan'
        Portifolio = pd.Series(Portifolio, index=self.AssetNames)
        Portifolio = Portifolio.to_frame().T.dropna(axis="columns")
        return Portifolio
