import pandas as pd
import numpy as np
import scipy.optimize as so
from ..TimeSeriesAnalysis import CovMatrix


class RiskParity():
    """
    Class to implement the risk parity portfolio

    Parameters
    ----------

    AssetPrices : :py.class:`pandas.DataFrame` with the asset prices

    RiskMetrics : string with the risk metrics to use.
        - var : Variance (default)
        - semi : Semi-variance

    Functions
    ---------

    Optimize : Function to optimize the weights of the portfolio using Risk Parity algorithm
    """

    def __init__(self, AssetPrices: pd.DataFrame, RiskMetrics: str = "var", **kwargs):

        # Check if the asset prices are valid
        if AssetPrices is not None:
            if not isinstance(AssetPrices, pd.DataFrame):
                raise ValueError(
                    "Asset Prices must be a Pandas DataFrame!")
            if not isinstance(AssetPrices.index, pd.DatetimeIndex):
                raise ValueError(
                    "Asset Prices must be a Pandas DataFrame index by datetime!")

        # Calculate the asset returns
        self.asset_returns_df = AssetPrices.pct_change().iloc[1:, :]

        # Get the number of assets and names
        self.asset_names = AssetPrices.columns
        self.num_assets = len(self.asset_names)

        if RiskMetrics.lower() in ["semi", "semivar", "semivariance"]:
            self.risk_matrix = CovMatrix.SampleSemi(AssetPrices)

        elif RiskMetrics.lower() in ["cov", "covariance", "variance", "var"]:
            self.risk_matrix = CovMatrix.Sample(AssetPrices)

        elif RiskMetrics.lower() in ["lw", "letoid", "letoidwolf"]:
            self.risk_matrix = CovMatrix.LedoitWolf(AssetPrices)

        else:
            raise ValueError(
                "RiskMetrics must be one 'semivariance' or 'variance'")

    @staticmethod
    def _getPortRisk(Weight, RiskMatrix):
        PortRisk = np.sqrt((Weight * RiskMatrix * Weight.T))[0, 0]
        return PortRisk

    @staticmethod
    def _AssetRiskContribution(Weight, RiskMatrix):
        PortRisk = RiskParity._getPortRisk(Weight, RiskMatrix)
        AssetRiskContribution = np.multiply(
            Weight.T, RiskMatrix * Weight.T) / PortRisk
        return AssetRiskContribution

    @staticmethod
    def _RiskContributionError(Weight, args):
        RiskMatrix = args[0]
        AssetRisk = args[1]
        Weight = np.matrix(Weight)
        PortRisk = RiskParity._getPortRisk(Weight, RiskMatrix)
        AssetRiskContribution = RiskParity._AssetRiskContribution(
            Weight, RiskMatrix)
        AssetRiskTarget = np.asmatrix(np.multiply(PortRisk, AssetRisk))
        Error = sum(np.square(AssetRiskContribution - AssetRiskTarget.T))[0, 0]
        return Error

    @staticmethod
    def _getRiskParity(RiskMatrix, AssetRisk, initial_Weight):
        constraints = ({'type': 'eq', 'fun': lambda x: np.sum(x) - 1.0},
                       {'type': 'ineq', 'fun': lambda x: x})

        optimize_result = so.minimize(fun=RiskParity._RiskContributionError,
                                      x0=initial_Weight,
                                      args=[RiskMatrix, AssetRisk],
                                      method='SLSQP',
                                      constraints=constraints,
                                      tol=1e-10,
                                      options={'disp': False})

        Weight = optimize_result["x"]
        return Weight

    def Optimize(self, Decimals=5) -> pd.DataFrame:
        """
        Function to optimize the weights of the portfolio using Risk Parity algorithm

        Parameters
        ----------

        Decimals : py:class:`int` with the number of decimals to round the weights

        Returns
        -------

        Portifolio : :py:class:`pandas.DataFrame` with the weights of the portfolio"""

        covariances = self.asset_returns_df.cov().values

        risk = [1 / self.num_assets] * self.num_assets

        start = [1 / self.num_assets] * self.num_assets

        Portifolio = RiskParity._getRiskParity(covariances, risk, start)
        Portifolio = Portifolio.round(Decimals)
        Portifolio /= Portifolio.sum()
        Portifolio[Portifolio == 0] = 'nan'

        Portifolio = pd.Series(Portifolio, index=self.asset_names)
        Portifolio = Portifolio.to_frame().T.dropna(axis="columns")
        return Portifolio
